#include "mcts_two_players.hpp"
#include "test_mcts_two_players.hpp"
#include "morpion.hpp"
#include <iostream>
#include <iomanip>
#include <map>

//#include <gperftools/profiler.h>

using namespace std;
using namespace game;

namespace mcts
{
  test_mcts_two_players::test_mcts_two_players() : openings_(morpion())
  {
    //self_play(1000);
    play();
    //test_openings(10000, 1000);
  }

  void test_mcts_two_players::test_openings(int nb_learning, int nb_testing)
  {
    //    self_play(nb_testing);
    self_play_learn_openings(nb_learning);
    self_play(nb_testing, true);
  }

  template <typename Game>
  int test_mcts_two_players::select_move(Game& game)
  {
    cout << game.player_to_string(game.current_player()) << " move: ";
    map<string, int> m;
    for (int i = 0; i < game.number_of_moves(); ++i)
      {
        m[game.move_to_string(i)] = i;
      }
    string move;
    getline(cin, move);
    game.play(m[move]);
    return m[move];
  }

  void test_mcts_two_players::play()
  {
    //    ProfilerStart("theturk.prof");
    morpion mor;
    auto the_turk = make_mcts_two_players(mor, 5000, 0.3, 4);
    cout << "play one game" << endl;
    cout << "who's first? (h)uman/(c)omputer ";
    string ans;
    getline(cin, ans);
    cout << mor << endl;
    int human_last_move = -1, computer_last_move = -1;
    while (!mor.end_of_game())
      {
        if ((ans == "h" && mor.current_player() == 0) || (ans == "c" && mor.current_player() == 1))
	  {
            human_last_move = select_move(mor);
	  }
        else
	  {
            if (human_last_move != -1 && computer_last_move != -1)
	      {
                the_turk.last_moves(computer_last_move, human_last_move);
	      }
            uint16_t move = the_turk.select_move();
            computer_last_move = move;
            cout << mor.player_to_string(mor.current_player()) << " move: " << mor.move_to_string(move) << endl;
            mor.play(move);
	  }
        cout << mor << endl;
      }
    if (mor.value(0) == 1) cout << mor.player_to_string(0) << " won";
    else if (mor.value(1) == 1) cout << mor.player_to_string(1) << " won";
    else cout << "draw";
    cout << endl;
    //    ProfilerStop();
  }

  void test_mcts_two_players::self_play_learn_openings(int n)
  {
    morpion mor;
    vector<uint16_t> moves(200);
    auto state = mor.get_state();
    auto the_turk_1 = make_mcts_two_players(mor, 1000, 0.6, 2);
    auto the_turk_2 = make_mcts_two_players(mor, 1000, 0.6, 2);
    map<set<int>, pair<uint32_t, double>> learning_examples;
    for (int i = 0; i < n; ++i)
      {
        cout << i << endl;
        cout << openings_ << endl << endl;
        moves.clear();
        mor.set_state(state);
        the_turk_1.reset();
        the_turk_1.init_with_openings(openings_);
        the_turk_2.reset();
        the_turk_2.init_with_openings(openings_);
        int the_turk_1_last_move = -1, the_turk_2_last_move = -1;
        int k = 0;
        while (!mor.end_of_game())
	  {
            if (k == 1) the_turk_2.last_move(the_turk_1_last_move);
            if (k % 2 == 0)
	      {
                if (the_turk_2_last_move != -1)
		  {
                    the_turk_1.last_moves(the_turk_1_last_move, the_turk_2_last_move);
		  }
                uint16_t move = the_turk_1.select_move();
                moves.push_back(move);
                the_turk_1_last_move = move;
                mor.play(move);
	      }
            else
	      {
                if (the_turk_2_last_move != -1)
		  {
                    the_turk_2.last_moves(the_turk_2_last_move, the_turk_1_last_move);
		  }
                uint16_t move = the_turk_2.select_move();
                moves.push_back(move);
                the_turk_2_last_move = move;
                mor.play(move);
	      }
            ++k;
	  }
	std::cout << mor << std::endl;
        int v = mor.value(0);
        cout << "value for first player " << v << endl;
        mor.set_state(state);
        openings_.update(mor, moves, v);
	mor.set_state(state);
	for (uint16_t m : moves) 
	  {
	    mor.play(m);
	    v = -v;
	    set<int> input_vector = std::move(mor.to_input_vector());
	    auto it = learning_examples.find(input_vector);
	    if (it == learning_examples.end())
	      {
		learning_examples[input_vector] = make_pair(1, v);
	      }
	    else
	      {
		it->second.second = (it->second.second * it->second.first + v) / (it->second.first + 1);
		it->second.first += 1;
	      }
	  }
      }
    cout << "number of learning examples: " << learning_examples.size() << endl;
    ofstream output("learning_examples.txt");
    for (const auto& example : learning_examples) 
      {
	output << example.second.second;
	for (int index : example.first) 
	  {
	    output << " " << index << ":" << 1;
	  }
	output << endl;
      }
    output.close();
  }

  void test_mcts_two_players::self_play(int n, bool with_openings)
  {
    morpion mor;
    auto state = mor.get_state();
    auto the_turk_v1 = make_mcts_two_players(mor, 1000, 0.6, 2);
    auto the_turk_v2 = make_mcts_two_players(mor, 1000, 0.6, 2);
    int nb_win_v1 = 0, nb_win_v2 = 0, nb_draw = 0;
    for (int i = 0; i < n; ++i)
      {
        cout << i << endl;
        mor.set_state(state);
        the_turk_v1.reset();
        the_turk_v2.reset();
	if (with_openings) the_turk_v2.init_with_openings(openings_);
        int the_turk_v1_last_move = -1, the_turk_v2_last_move = -1;
	int k = 0;
        while (!mor.end_of_game())
	  {
            if (with_openings && k == 1 && i % 2 == 0) the_turk_v2.last_move(the_turk_v1_last_move);
	    ++k;
            //	    cout << mor << endl;
            if ((i % 2 == 0 && mor.current_player() == 0) || (i % 2 == 1 && mor.current_player() == 1))
	      {
                if (the_turk_v1_last_move != -1 && the_turk_v2_last_move != -1)
		  {
                    the_turk_v1.last_moves(the_turk_v1_last_move, the_turk_v2_last_move);
		  }
                uint16_t move = the_turk_v1.select_move();
                the_turk_v1_last_move = move;
                mor.play(move);
	      }
            else
	      {
                if (the_turk_v1_last_move != -1 && the_turk_v2_last_move != -1)
		  {
                    the_turk_v2.last_moves(the_turk_v2_last_move, the_turk_v1_last_move);
		  }
                uint16_t move = the_turk_v2.select_move();
                the_turk_v2_last_move = move;
                mor.play(move);
	      }
	  }
        if (mor.value(0) == 1)
	  {
            if (i % 2 == 0)
	      {
                /*cout << "v1 won" << endl;*/ ++nb_win_v1;
	      }
            else
	      {
                /*cout << "v2 won" << endl;*/ ++nb_win_v2;
	      }
	  }
        else if (mor.value(1) == 1)
	  {
            if (i % 2 == 0)
	      {
                /*cout << "v2 won" << endl;*/ ++nb_win_v2;
	      }
            else
	      {
                /*cout << "v1 won" << endl;*/ ++nb_win_v1;
	      }
	  }
        else
	  {
            /*cout << "draw" << endl;*/ ++nb_draw;
	  }
        cout << setw(10) << "v1 nb wins: " << nb_win_v1 << " v2 nb wins: " << nb_win_v2 << " nb draws: " << nb_draw << endl;
      }
  }

  void test_mcts_two_players::self_play()
  {
    morpion mor;
    auto the_turk_v1 = make_mcts_two_players(mor, 1000, 1.2, 2);
    auto the_turk_v2 = make_mcts_two_players(mor, 1000, 1.2, 2);
    cout << "play one game" << endl;
    cout << "who's first? the_turk_(v1)/the_turk_(v2) ";
    string ans;
    getline(cin, ans);
    cout << mor << endl;
    int the_turk_v1_last_move = -1, the_turk_v2_last_move = -1;
    while (!mor.end_of_game())
      {
        if ((ans == "v1" && mor.current_player() == 0) || (ans == "v2" && mor.current_player() == 1))
	  {
            if (the_turk_v1_last_move != -1 && the_turk_v2_last_move != -1)
	      {
                the_turk_v1.last_moves(the_turk_v1_last_move, the_turk_v2_last_move);
	      }
            uint16_t move = the_turk_v1.select_move();
            the_turk_v1_last_move = move;
            cout << mor.player_to_string(mor.current_player()) << " move: " << mor.move_to_string(move) << endl;
            mor.play(move);
	  }
        else
	  {
            if (the_turk_v1_last_move != -1 && the_turk_v2_last_move != -1)
	      {
                the_turk_v2.last_moves(the_turk_v2_last_move, the_turk_v1_last_move);
	      }
            uint16_t move = the_turk_v2.select_move();
            the_turk_v2_last_move = move;
            cout << mor.player_to_string(mor.current_player()) << " move: " << mor.move_to_string(move) << endl;
            mor.play(move);
	  }
        cout << mor << endl;
      }
    if (mor.value(0) == 1) cout << mor.player_to_string(0) << " won";
    else if (mor.value(1) == 1) cout << mor.player_to_string(1) << " won";
    else cout << "draw";
    cout << endl;
  }

}
